# RemoteKB
#
# Copyright (c) 2008 Arkadiusz Wahlig
# <arkadiusz.wahlig@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import e32, sys, os, appuifw, socket, types, time
from key_codes import *

# try to import the keypress module
try:
    import keypress
except ImportError:
    appuifw.note(u'Requires keypress 1.0.6+ module', 'error')
    appuifw.app.set_exit()
else:
    if not hasattr(keypress, 'version_info') or keypress.version_info < (1, 0, 5):
        appuifw.note(u'Requires keypress 1.0.6+ module', 'error')
        appuifw.app.set_exit()


__version__ = '1.00.2'


def bt_discover_service():
    try:
        addr, services = socket.bt_discover()
    except socket.error, err:
        if err[0] == 2: # "no such file or directory"
            appuifw.note(u'No serial ports found', 'error')
        elif err[0] == 13: # "permission denied"
            appuifw.note(u'Permission denied', 'error')
        elif err[0] != 4: # "interrupted system call"
            raise
        return
    if not services:
        # may this happen?
        return
    if len(services) > 1:
        choices = services.keys()
        choices.sort()
        def formatname(name):
            # try to shorten the name
            for s in (u'Bluetooth', u'bluetooth', u'\x00'):
                name = name.replace(s, u'').strip()
            return name
        n = appuifw.popup_menu([u'%s. %s' % \
            (services[x], formatname(x)) for x in choices],
            u'Choose port')
        if n is None:
            return
        port = services[choices[n]]
    else:
        port = services.values()[0]
    return addr, port


class BT_Keyboard(object):
    TYPE_SIMULATE_KEY = 0
    TYPE_SIMULATE_EVENT_UP = 1
    TYPE_SIMULATE_EVENT_DOWN = 2
    TYPE_SIMULATE_EVENT_KEY = 3
    TYPE_SEND_RAW_EVENT_UP = 4
    TYPE_SEND_RAW_EVENT_DOWN = 5
    
    def __init__(self, **events):
        self.sock = socket.socket(socket.AF_BT, socket.SOCK_STREAM, 0, None)
        self.recvbuf = ''
        self.events = events

    def connect(self, target):
        if self.sock is None:
            raise RuntimeError('closed')
        self.sock.connect(target)
        self.set_recv_callback()

    def close(self):
        try:
            self.events.clear()
            # NOTE: shutdown will cause an error in recv_callback
            self.sock.shutdown(2)
            self.sock = None
        except:
            pass

    def set_recv_callback(self):
        self.sock.recv(128, 0, self.recv_callback)

    def recv_callback(self, data):
        if isinstance(data, types.StringType):
            if data:
                self.recvbuf += data
                try:
                    self.handle_keys()
                except Exception, e:
                    from globalui import global_note
                    global_note(u'RemoteKB: %s' % e, 'error')
            self.set_recv_callback()
        else:
            self.call_event('event_socket_error', data[-1])
            self.sock = None

    def handle_keys(self):
        while self.recvbuf:
            pos = self.recvbuf.find('\r')
            if pos < 0:
                break
            data, self.recvbuf = self.recvbuf[:pos], self.recvbuf[pos+1:]

            try:
                o = eval(data)
                type = o[0]
                args = o[1:]

                if type == self.TYPE_SIMULATE_KEY:
                    e32.reset_inactivity()
                    keypress.simulate_key(*args)
                elif type == self.TYPE_SIMULATE_EVENT_UP:
                    keypress.simulate_event(keypress.EEventKeyUp, *args)
                elif type == self.TYPE_SIMULATE_EVENT_DOWN:
                    e32.reset_inactivity()
                    keypress.simulate_event(keypress.EEventKeyDown, *args)
                elif type == self.TYPE_SIMULATE_EVENT_KEY:
                    e32.reset_inactivity()
                    keypress.simulate_event(keypress.EEventKey, *args)
                elif type == self.TYPE_SEND_RAW_EVENT_UP:
                    keypress.send_raw_event(keypress.EKeyUp, *args)
                elif type == self.TYPE_SEND_RAW_EVENT_DOWN:
                    keypress.send_raw_event(keypress.EKeyDown, *args)
            except Exception, e:
                from globalui import global_note
                global_note(u'RemoteKB: %s' % e, 'error')
                continue

    def call_event(self, name, *args, **kwargs):
        event = self.events.get(name, None)
        if event:
            return event(*args, **kwargs)
    

class Application(object):
    def __init__(self):
        self.kb = None
        self.lbox = None
        self.target = None
        self.busy = False
        
    def connect(self):
        self.menu[1][1] = u'Connecting...'
        self.update()
        reset_status = True

        self.busy = True
        try:
            target = bt_discover_service()
            if target:
                self.kb = BT_Keyboard(event_socket_error=self.event_socket_error)
                try:
                    self.kb.connect(target)
                except:
                    self.kb = None
                    appuifw.note(u'Failed to connect', 'error')
                else:
                    self.menu[0] = (u'Disconnect', u'Disconnect from a PC', self.disconnect)
                    self.menu[1][1] = u'Connected to %s/%d' % target
                    self.update()
                    reset_status = False
                    appuifw.note(u'Connected', 'conf')
        finally:
            if reset_status:
                self.menu[1][1] = u'Not connected'
            self.update()
            self.busy = False

    def disconnect(self):
        if not appuifw.query(u'Sure to disconnect?', 'query'):
            return

        self.kb.close()
        self.kb = None

        self.menu[0] = self.startmenu[0]
        self.menu[1][1] = u'Not connected'
        self.update()

    def status(self):
        appuifw.note(self.menu[self.lbox.current()][1])

    def select(self):
        if self.busy:
            return
        try:
            target = self.menu[self.lbox.current()][2]
            if target:
                target()
        except SystemExit:
            raise
        except Exception, e:
            appuifw.note(unicode(e), 'error')

    def start(self):
        appuifw.app.menu = [(u'Select', self.select),
            (u'Exit', self.stop)]

        appuifw.app.exit_key_handler = self.stop
    
        self.menu = [(u'Connect', u'Connect to a PC over Bluetooth', self.connect),
            [u'Status', u'Not connected', self.status],
            (u'Help', u'Display help', self.help),
            (u'Exit', u'Close application', self.stop)]
        self.startmenu = list(self.menu)
    
        self.lbox = appuifw.Listbox([(a, b) for a, b, c in self.menu], self.select)
        appuifw.app.body = self.lbox
    
    def stop(self):
        if self.kb:
            if not appuifw.query(u'Sure to disconnect and exit?', 'query'):
                return

        appuifw.app.set_exit()
        sys.exit()

    def help(self):
        try:
            path = os.path.split(__file__)[0]
            f = open(os.path.join(path, 'help.txt'), 'r')
            help = f.read().decode('utf8')
            f.close()
        except IOError:
            appuifw.note(u'Could not load the help file', 'error')
            return
        lock = e32.Ao_lock()
        help = help.replace(u'{version}', __version__, 1)
        state = appuifw.app.menu, appuifw.app.body, appuifw.app.exit_key_handler
        text = appuifw.Text()
        text.color = 0x0000ff
        text.add(help)
        text.set_pos(0)
        appuifw.app.menu = [(u'Exit', lock.signal)]
        appuifw.app.exit_key_handler = lock.signal
        appuifw.app.body = text
        lock.wait()
        appuifw.app.menu, appuifw.app.body, appuifw.app.exit_key_handler = state

    def update(self):
        self.lbox.set_list([(a, b) for a, b, c in self.menu], self.lbox.current())
        e32.ao_yield()
        
    def event_socket_error(self, error):
        appuifw.note(u'Disconnected')
        self.kb = None
        self.menu[0] = self.startmenu[0]
        self.menu[1][1] = u'Not connected'
        self.update()
    

def main():
    # try to make us a system app if envy module is available
    try:
        from envy import set_app_system
        set_app_system(1)
    except:
        pass
        
    global app
    app = Application()
    app.start()
